/*!
 * \file
 * \brief   Distributed bitonic implementation.
 *
 * \author
 *    Christos Choutouridis AEM:8997
 *    <cchoutou@ece.auth.gr>
 */

#include <vector>
#include <algorithm>
#include <cmath>
#if !defined DEBUG
#define NDEBUG
#endif
#include <cassert>

#include "distbitonic.hpp"



/*!
 * Returns the ascending or descending configuration of the node's sequence based on
 * the current node (MPI process) and the depth of the sorting network
 *
 * @param node      The current node (MPI process)
 * @return          True if we need ascending configuration, false otherwise
 */
template <>
bool ascending<SortMode::Bubbletonic>(mpi_id_t node, [[maybe_unused]] size_t depth) noexcept {
    return (node % 2) == 0;
}

/*!
 * Returns the ascending or descending configuration of the node's sequence based on
 * the current node (MPI process) and the depth of the sorting network
 *
 * @param node      The current node (MPI process)
 * @param depth     The total depth of the sorting network (same for each step for a given network)
 *
 * @return          True if we need ascending configuration, false otherwise
 */
template <>
bool ascending<SortMode::Bitonic>(mpi_id_t node, size_t depth) noexcept {
    return !(node & (1 << depth));
}

/*!
 * Returns the node's partner for data exchange during the sorting network iterations
 * of Bubbletonic
 *
 * @param node      The current node
 * @param step      The step of the sorting network
 * @return          The node id of the partner for data exchange
 */
template <>
mpi_id_t partner<SortMode::Bubbletonic>(mpi_id_t node, size_t step) noexcept {
//    return (node % 2 == step % 2) ? node + 1 : node - 1;
    return (((node+step) % 2) == 0) ? node + 1 : node - 1;
}

/*!
 * Returns the node's partner for data exchange during the sorting network iterations
 * of Bitonic
 *
 * @param node      The current node
 * @param step      The step of the sorting network
 * @return          The node id of the partner for data exchange
 */
template <>
mpi_id_t partner<SortMode::Bitonic>(mpi_id_t node, size_t step) noexcept {
    return (node ^ (1 << step));
}


/*!
 * Predicate to check if a node keeps the small numbers during the bubbletonic sort network exchange.
 *
 * @param node      The node for which we check
 * @param partner   The partner of the data exchange
 * @return          True if the node should keep the small values, false otherwise
 */
template <>
bool keepSmall<SortMode::Bubbletonic>(mpi_id_t node, mpi_id_t partner, [[maybe_unused]] size_t depth) noexcept {
    assert(node != partner);
    return (node < partner);
}

/*!
 * Predicate to check if a node keeps the small numbers during the bitonic sort network exchange.
 *
 * @param node      The node for which we check
 * @param partner   The partner of the data exchange
 * @param depth     The total depth of the sorting network (same for each step for a given network)
 * @return          True if the node should keep the small values, false otherwise
 */
template <>
bool keepSmall<SortMode::Bitonic>(mpi_id_t node, mpi_id_t partner, size_t depth) noexcept {
    assert(node != partner);
    return ascending<SortMode::Bitonic>(node, depth) == (node < partner);
}

/*!
 * Predicate to check if the node is active in the current iteration of the bubbletonic
 * sort exchange.
 *
 * @param node      The node to check
 * @param nodes     The total number of nodes
 * @return          True if the node is active, false otherwise
 */
bool isActive(mpi_id_t node, mpi_id_t nodes) noexcept {
    return (node >= 0) && (node < (nodes-1));
}


void exchange(mpi_id_t node, mpi_id_t partner) {
    assert(node != partner);
}

void minmax(AllData_t& data, mpi_id_t node, mpi_id_t partner, bool keepsmall) {
    for (size_t i = 0; i < data[node].size(); ++i) {
        if (keepsmall && data[node][i] > data[partner][i])
            std::swap(data[node][i], data[partner][i]);
        if (!keepsmall && data[node][i] < data[partner][i])
            std::swap(data[node][i], data[partner][i]);
    }
}




void bubbletonic_network(AllData_t& data, mpi_id_t nodes, size_t depth) {
    for (mpi_id_t node = 0 ; node < nodes ; ++node) { // Currently we do all nodes here!
        auto part = partner<SortMode::Bubbletonic>(node, depth);
        auto ks = keepSmall<SortMode::Bubbletonic>(node, part, 0);
        if (isActive(node, nodes) && node < part) {
            exchange(node, part);
            minmax(data, node, part, ks);
            // elbow-sort here
            if (ascending<SortMode::Bubbletonic>(node, 0))
                std::sort(data[node].begin(), data[node].end(), std::less<>());
            else
                std::sort(data[node].begin(), data[node].end(), std::greater<>());

            if (ascending<SortMode::Bubbletonic>(part, 0))
                std::sort(data[part].begin(), data[part].end(), std::less<>());
            else
                std::sort(data[part].begin(), data[part].end(), std::greater<>());
        }
    }
}

void distBubbletonic(mpi_id_t P, AllData_t& data) {
    for (mpi_id_t node = 0 ; node < P ; ++node) { // Currently we do all nodes here!
        // Initially sort to create the half part of a bitonic
        if (ascending<SortMode::Bubbletonic>(node, 0))
            std::sort(data[node].begin(), data[node].end(), std::less<>());
        else
            std::sort(data[node].begin(), data[node].end(), std::greater<>());
    }

    for (size_t depth = 0; depth < P-1; ++depth) {
        bubbletonic_network(data, P, depth);
    }

    // Invert the descending ones
    for (mpi_id_t node = 0 ; node < P ; ++node) { // Currently we do all nodes here!
        if (!ascending<SortMode::Bubbletonic>(node, 0))
            std::sort(data[node].begin(), data[node].end(), std::less<>());
    }
}


void bitonic_network(AllData_t& data, mpi_id_t nodes, size_t depth) {
    for (size_t step = depth; step > 0;) {
        --step;
        for (mpi_id_t node = 0; node < nodes; ++node) { // Currently we do all nodes here!
            auto part = partner<SortMode::Bitonic>(node, step);
            auto ks = keepSmall<SortMode::Bitonic>(node, part, depth);
            if (node < part) {
                exchange(node, part);
                minmax(data, node, part, ks);
            }
        }
    }
}

void distBitonic(mpi_id_t P, AllData_t& data) {
    auto p = static_cast<uint32_t>(std::log2(P));

    for (mpi_id_t node = 0 ; node < P ; ++node) { // Currently we do all nodes here!
        // Initially sort to create the half part of a bitonic
        if (ascending<SortMode::Bitonic>(node, 0))
            std::sort(data[node].begin(), data[node].end(), std::less<>());
        else
            std::sort(data[node].begin(), data[node].end(), std::greater<>());
    }

    // Run through sort network using elbow-sort
    for (size_t depth = 1; depth <= p; ++depth) {
        bitonic_network(data, P, depth);

        for (mpi_id_t node = 0 ; node < P ; ++node) { // Currently we do all nodes here!
            // elbow-sort here
            if (ascending<SortMode::Bitonic>(node, depth))
                std::sort(data[node].begin(), data[node].end(), std::less<>());
            else
                std::sort(data[node].begin(), data[node].end(), std::greater<>());
        }

    }
}
